from django.shortcuts import render, redirect
from django.views.generic import TemplateView
from django.http import HttpResponseRedirect
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.contrib.auth import authenticate, login as auth_login
from django.contrib.auth.decorators import login_required
from django.contrib import messages
from .models import *
from .forms import *
from .filters import *
from .decorators import *
import numpy


# HELPERS
# I think eventually we should change this
# If a user creates a product called "This is my product,"
# then the URL would be "/product/This is my product" which
# is very gross. Perhaps add a __str__ function to the Model
# that returns all lowercase string of first word in name?
def retrieve_product(product_name):
    """
    Queries DB for a product by name (This will be updated to query by ID)

    :param product_name: The name of target product (This will be updated to query by ID)
    :type request: str

    :return: a Product object on success, None on failure
    :rtype: Product or None
    """
    try:
        return Product.objects.get(name=product_name)
    except:
        return None


def retrieveDependencies(product):
    try:
        return Dependency.objects.filter(dependent=product)
    except:
        return []


# TODO: could replace this with .exists()?
def retrieveProductError(_product):
    try:
        return DependencyCycleError.objects.get(product_id=_product.id)
    except:
        return None


# def handle_edit_product(form, name):

def generate_paginator(current_page_index, last_page):
    """
    Determines the paginator range values

    :param current_page_index: The currently viewed page (indexed, so page - 1)
    :type current_page_index: unsigned integer

    :param last_page: last page in the paginator (indexed, so len - 1)
    :type last_page: unsigned integer

    :return: an array of integers, separated with a tuple ('...', integer) to indicate a break in the list 
        and the page to jump to on a click of '...'
    :rtype: Array
    """
    page_range = []

    if current_page_index - 3 >= 0 and last_page - 3 > current_page_index:
        page_range.append(1)
        page_range.append(('...', current_page_index - 1))
        for i in range(current_page_index - 1, current_page_index + 2):
            page_range.append(i + 1)
        page_range.append(('...', current_page_index + 3))

        for i in range(last_page - 2, last_page):
            page_range.append(i + 1)

    # elif current_page_index - 3 < 0:
    else:
        for i in range(0, 3):
            page_range.append(i + 1)
        page_range.append(('...', 4))

        for i in range(last_page - 3, last_page):
            page_range.append(i + 1)

    return page_range


### VIEWS ###
@unauthed_route
def login(request):
    """
    Logs a user in

    :param request: The request sent to server
    :type request: HttpRequest

    :return: a redirect to the product list page on sucess, remains on this page on failure
    :rtype: HttpResponseRedirect
    """
    if request.method == "POST":
        username = request.POST.get("username")
        password = request.POST.get("password")

        user = authenticate(request, username=username, password=password)

        if user is not None:
            auth_login(request, user)
            return redirect('/products/')
        else:
            messages.info(request, "Username or password is incorrect")

    return render(request, 'home/login.html')


@unauthed_route
def register(request):
    """
    Registers a new user

    :param request: The request sent to server
    :type request: HttpRequest

    :return: a redirect to the login page on sucess, remains on this page on failure
    :rtype: HttpResponseRedirect
    """
    if request.method == "POST":
        register_form = CreateUserForm(request.POST)

        if register_form.is_valid():
            register_form.save()
            messages.success(request, 'Account was created successfully')
            return redirect("/login")
        else:
            print(register_form.errors)
            for message in list(register_form.errors.values()):
                messages.info(request, message)

    return render(request, 'home/register.html')


def fourohfour(request):
    """
    GET request handler for the URL '/fourohfour'

    :param request: The request sent to server
    :type request: HttpRequest

    :return: HttpResponse containing the page HTML, an HttpResponseRedirect to the 404 page '/fourohfour' if 
        request type isn't a GET
    :rtype: HttpResponse or HttpResponseRedirect
    """
    return render(request, 'fourohfour/fourohfour.html')


def home(request):
    """
    GET request handler for the URL '/'

    :param request: The request sent to server
    :type request: HttpRequest
    
    :return: HttpResponse containing the page HTML
    :rtype: HttpResponse
    """
    return render(request, 'home/index.html')


# @login_required(login_url='/login/')
def errors_page(request):
    if request.method != 'GET':
        return HttpResponseRedirect("/fourohfour")

    error_list = []
    for p in DependencyCycleError.objects.all():
        error_list.append(p.product.id)
    product_list = Product.objects.filter(id__in=error_list)

    context = {
        'products': product_list,
    }

    return render(request, 'product_pages/errorlisting.html', context)


def products_page(request):
    """
    GET request handler for the URL '/'

    :param request: The request sent to server
    :type request: HttpRequest
    
    :return: HttpResponse containing the page HTML and a context of a list of Product objects, and an
         HttpResponseRedirect to the 404 page '/fourohfour' if request type isn't a GET
    :rtype: HttpResponse or HttpResponseRedirect
    """
    if request.method != 'GET':
        return HttpResponseRedirect("/fourohfour")

    page = request.GET.get('page', 1)

    has_filter = request.GET.get('note')

    # this is not necessary, this is just to keep consistency
    # that /products alone defaults to page 1
    if page is not None and page == '1':
        return HttpResponseRedirect("/products")

    if has_filter == '':
        return HttpResponseRedirect("/products/")

    product_list = Product.objects.all()


    
    myFilter = ProductFilter(request.GET, queryset=product_list) # Instantiates filter using definiton from filters.py
    product_list = myFilter.qs                                   # Creates a query set by filtering the data


    paginator = Paginator(product_list, 10)
    errors_exist = DependencyCycleError.objects.exists()
    # print(errors_exist)

    try:
        products = paginator.page(page)
    except PageNotAnInteger:
        products = paginator.page(1)
        page = 1
    except EmptyPage:
        products = paginator.page(paginator.num_pages)
        page = paginator.num_pages

    page_range = generate_paginator(products.number - 1, len(paginator.page_range))

    context = {
        'products': products,
        'myFilter': myFilter,
        'has_filter': has_filter,
        'current_page': page,
        'errors': errors_exist,
        'page_range': page_range
    }

    return render(request, 'product_pages/index.html', context)


# TODO:
# 1). Add modals for creating / deleting dependencies
# 2). Add Button + Confirmation modal (i.e. "Are you sure?") for delete product

def product_view(request, name):
    """
    GET request handler for the URL '/product/:id'

    :param request: The request sent to server
    :type request: HttpRequest

    :param name: the name of the product selected
    :type name: str (This will change to int to reflect the future change in the Product model)

    :return: HttpResponse containing the page HTML and a context of the selected Product object, and an
         HttpResponseRedirect to the 404 page '/fourohfour' if request type isn't a GET
    :rtype: HttpResponse or HttpResponseRedirect
    """
    target_product = retrieve_product(name)
    product_dependencies = retrieveDependencies(target_product)
    graph_error = retrieveProductError(target_product)

    product_list = Product.objects.all()
    

    if target_product is None or request.method != 'GET':
        return redirect('/fourohfour')
    else:
        context = {
            'product': target_product,
            'dependencies': product_dependencies,
            'products': product_list,
            'graph_error': graph_error
        }
        return render(request, 'product_pages/product_info.html', context)


def product_analytics(request, name):
    """
    GET request handler for the URL '/product/:id/analytics/'

    :param request: The request sent to server
    :type request: HttpRequest

    :param name: the name of the product selected
    :type name: str (This will change to int to reflect the future change in the Product model)

    :return: HttpResponse containing the page HTML and a context of the selected Product object, and an
         HttpResponseRedirect to the 404 page '/fourohfour' if request type isn't a GET
    :rtype: HttpResponse or HttpResponseRedirect
    """
    target_product = retrieve_product(name)

    if target_product is None or request.method != "GET":
        return redirect('/fourohfour')
    else:
        labels = []
        real_price = []
        estimated_labor_time = []
        estimated_cost = []

        all_history = ProductHistory.objects.all()
        j = 1
        for i in range(0, len(all_history)):
            if name == all_history[i].name:
                labels.append(f"{all_history[i].history_point.date_time:%Y/%m/%d %H:%M}")
                real_price.append(all_history[i].real_price)
                estimated_labor_time.append(all_history[i].direct_labor + all_history[i].indirect_labor)
                estimated_cost.append(all_history[i].indirect_wages + all_history[i].direct_wages)
        cor_labor_price = numpy.corrcoef(numpy.array(real_price), numpy.array(estimated_labor_time))
        cor_labor_price = cor_labor_price[0,1]
        cor_cost_price = numpy.corrcoef(estimated_cost, real_price)
        cor_cost_price = cor_cost_price[0,1]

        chartjs_config = {
            'type': 'line',
            'data': {
                'labels': labels,
                'datasets': [
                    {
                        'label': "Real Price",
                        'borderColor': "rgba(0,33,165,1)",
                        'yAxisID': 'l',
                        'data': real_price
                    },
                    {
                        'label': "Estimated Cost",
                        'borderColor': "rgba(250,70,22,1)",
                        'yAxisID': "l",
                        'data': estimated_cost
                    },
                    {
                        'label': "Estimated Labor Time",
                        'borderColor': 'rgba(25, 25, 25, 1)',
                        'pointHoverBorderColor': 'rgba(25, 25, 25, 1)',
                        'yAxisID': "r",
                        'data': estimated_labor_time
                    }
                ]},
            'options': {
                'responsive': 'true',
                'legend': {
                    'position': 'top',
                    'display': 'true'
                    },
                'scales': {
                    'yAxes': [{
                        'id': "l",
                        'type': "linear",
                        'position': "left",
                        'scaleLabel': {
                            'display': 'true',
                            'labelString': 'Dollars ($)'}
                    }, {
                        'id': "r",
                        'type': "linear",
                        'position': "right",
                        'scaleLabel': {
                            'display': 'true',
                            'labelString': 'Hours'}
                    }]
                }
            }

        }

        context = {
            'product': target_product,
            'dependencies': [],
            'labels': labels,
            'chartjs_config': chartjs_config,
            'cor_labor_price': cor_labor_price,
            'cor_cost_price': cor_cost_price

        }
        return render(request, 'product_pages/product_analytics.html', context)


def products_analytics(request):
    """
    GET request handler for the URL '/products/analytics/'

    :param request: The request sent to server
    :type request: HttpRequest

    :return: HttpResponse containing the page HTML and a context of the Product objects, and an
         HttpResponseRedirect to the 404 page '/fourohfour' if request type isn't a GET
    :rtype: HttpResponse or HttpResponseRedirect
    """
    if request.method != "GET":
        return redirect('/fourohfour')

    labels = []
    real_price = []
    estimated_labor_time = []
    estimated_cost = []

    all_prods = Product.objects.all().order_by('real_price')
    j = 1
    for prod in all_prods:
        labels.append(f"{prod.name}")
        real_price.append(prod.real_price)
        estimated_labor_time.append(prod.value)
        estimated_cost.append(prod.cost_price)
    cor_labor_price = numpy.corrcoef(numpy.array(real_price), numpy.array(estimated_labor_time))
    cor_labor_price = cor_labor_price[0,1]
    cor_cost_price = numpy.corrcoef(estimated_cost, real_price)
    cor_cost_price = cor_cost_price[0,1]

    chartjs_config = {
        'type': 'line',
        'data': {
            'labels': labels,
            'datasets': [
                {
                    'label': "Real Price",
                    'borderColor': "rgba(0,33,165,1)",
                    'yAxisID': 'l',
                    'data': real_price
                },
                {
                    'label': "Estimated Cost",
                    'borderColor': "rgba(250,70,22,1)",
                    'yAxisID': "l",
                    'data': estimated_cost
                },
                {
                    'label': "Estimated Labor Time",
                    'borderColor': 'rgba(25, 25, 25, 1)',
                    'pointHoverBorderColor': 'rgba(25, 25, 25, 1)',
                    'yAxisID': "r",
                    'data': estimated_labor_time
                }
            ]},
        'options': {
            'responsive': 'true',
            'legend': {
                'position': 'top',
                'display': 'true'
                },
            'scales': {
                'yAxes': [{
                    'id': "l",
                    'type': "linear",
                    'position': "left",
                    'scaleLabel': {
                        'display': 'true',
                        'labelString': 'Dollars ($)'}
                }, {
                    'id': "r",
                    'type': "linear",
                    'position': "right",
                    'scaleLabel': {
                        'display': 'true',
                        'labelString': 'Hours'}
                }]
            }
        }

    }

    context = {
        'dependencies': [],
        'labels': labels,
        'chartjs_config': chartjs_config,
        'cor_labor_price': cor_labor_price,
        'cor_cost_price': cor_cost_price

    }
    return render(request, 'product_pages/products_analytics.html', context)




from django.test import TestCase
from ..models import *
from ..api import *

class HistoryTestCase(TestCase):
    def setUp(self):
        p1 = Product(name="Prod1", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p2 = Product(name="Prod2", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p3 = Product(name="Prod3", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p4 = Product(name="Prod4", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p1.save()
        p2.save()
        p3.save()
        p4.save()
        d1 = Dependency(dependent_id=p2.id, dependency_id=p1.id, quantity =1.0)
        d2 = Dependency(dependent_id=p3.id, dependency_id=p1.id, quantity =1.0)
        d3 = Dependency(dependent_id=p3.id, dependency_id=p2.id, quantity =1.0)
        d4 = Dependency(dependent_id=p4.id, dependency_id=p3.id, quantity =1.0)
        d1.save()
        d2.save()
        d3.save()
        d4.save()

    def test_history_logs_one_commit(self):
        update_product_indirect_values()
        commit_history()
        self.assertEqual(ProductHistory.objects.get(product_id=1, history_point_id=1).indirect_labor, 0.0)
        self.assertEqual(ProductHistory.objects.get(product_id=2, history_point_id=1).indirect_labor, 10.0)
        self.assertEqual(ProductHistory.objects.get(product_id=3, history_point_id=1).indirect_labor, 30.0)
        self.assertEqual(ProductHistory.objects.get(product_id=4, history_point_id=1).indirect_labor, 40.0)

    def test_history_logs_two_commits(self):
        update_product_indirect_values()
        commit_history()
        Product.objects.filter(id=3).update(direct_labor=20.0)
        update_product_indirect_values()
        commit_history()
        self.assertEqual(ProductHistory.objects.get(product_id=4, history_point_id=1).indirect_labor, 40.0)
        self.assertEqual(ProductHistory.objects.get(product_id=4, history_point_id=2).indirect_labor, 50.0)




from django.db import models
import uuid

# TODO: Normalize? 
class Product(models.Model):
    """This class is a model for data that relates directly to products in the economy."""
    name = models.CharField(max_length = 100, unique = True) # NOTE: this max length is arbitrary placeholder
    # The unit of measurement for a product
    measurement = models.CharField(max_length = 40)
    # The price of the product in the real world
    real_price = models.FloatField()
    # The labor hours directly expended in production of the product
    direct_labor = models.FloatField() # Consider placing a lower bound of 0.0?
    # The wages directly expended in production of the product
    direct_wages = models.FloatField()
    # The wages expended in production of the product's dependencies
    indirect_wages = models.FloatField(default=0)
    # The labor hours expended in production of the product's dependencies
    indirect_labor = models.FloatField(default=0)

    @property
    def cost_price(self):
        # wage price + wage price of deps + profits of deps. NOTE: very rough estimate of profits, hardcoded
        return self.direct_wages + self.indirect_wages + (self.indirect_wages * 0.1)

    @property
    def value(self):
        return self.direct_labor + self.indirect_labor

class Dependency(models.Model):
    """This class is a model for dependency relations between different products.
    The dependent depends on a certain quantity of the dependency to be produced."""
    dependent = models.ForeignKey(
            Product,
            on_delete=models.CASCADE,
            related_name = 'dependents'
            )
    dependency = models.ForeignKey(
            Product,
            on_delete=models.CASCADE,
            related_name = 'dependencies'
            )
    quantity = models.FloatField()

class DependencyCycleError(models.Model):
    """This class is a model used to list Products that depend on 1 or more of themselves,
    or on another product which does so."""
    product = models.ForeignKey(
            Product,
            on_delete=models.CASCADE,
            )

# This is currently unused
class Economy(models.Model):
    real_average_profit_rate = models.FloatField()
    estimated_average_profit_rate = models.FloatField()

class HistoryPoint(models.Model):
    """This class is a model that is used for records of past states of the Product graph."""
    date_time = models.DateTimeField(auto_now_add=True)
    # TODO: Other fields worth recording here? Profit rate? other Model for storing some calculated values?

class ProductHistory(models.Model):
    """This class is a model that stores historical Product data."""
    history_point = models.ForeignKey(
            HistoryPoint,
            on_delete = models.CASCADE
            )
    product_id = models.IntegerField() # NOTE: This is not a foreign key field because reasons
    name = models.CharField(max_length = 100) # NOTE: this max length is arbitrary placeholder
    measurement = models.CharField(max_length = 40)
    real_price = models.FloatField()
    direct_labor = models.FloatField() # Consider placing a lower bound of 0.0?
    direct_wages = models.FloatField()
    indirect_wages = models.FloatField(default=0)
    indirect_labor = models.FloatField(default=0)

    @property
    def cost_price(self):
        # wage price + wage price of deps + profits of deps. NOTE: very rough estimate of profits, hardcoded
        return self.direct_wages + self.indirect_wages + (self.indirect_wages * 0.1)

    @property
    def value(self):
        return self.direct_labor + self.indirect_labor

# NOTE: actually displaying the history for these is not a priority
class DependencyHistory(models.Model):
    """This class is a model that stores historical Dependency data."""
    history_point = models.ForeignKey(
            HistoryPoint,
            on_delete = models.CASCADE
            )
    dependent_id = models.IntegerField()
    dependency_id = models.IntegerField()
    quantity = models.FloatField()



import django_filters 
from .models import *
from django_filters import CharFilter

class ProductFilter(django_filters.FilterSet):
    note = CharFilter(field_name='name', lookup_expr="icontains")

    class Meta:
        model = Product
        fields = ('__all__')
        exclude = ['real_price', 'direct_labor', 'direct_wages', 'indirect_wages', 'indirect_labor', 'measurement']


from django.test import Client
from django.test import TestCase

class TestGetURLs(TestCase):
    def setUp(self):
        self.client = Client()

    def test_home(self):
        #Get Response to Home
        res = self.client.get('/')
        self.assertEqual(res.status_code, 200)

    def test_admin(self):
        #Get Response to Admin
        res = self.client.get('/admin')
        self.assertEqual(res.status_code, 301)

        #Get Response to correct Admin
        res = self.client.get('/admin/login/?next=/admin/')
        self.assertEqual(res.status_code, 200)

    def test_fourohfour(self):
        #Get Request to fourohfour
        response = self.client.get('/fourohfour')
        self.assertEqual(response.status_code, 301)

class TestPostURLS(TestCase):
    def setUp(self):
        self.client = Client()
        self.client.post("/api/create/product", {
            'name': 'borgor', 'measurement': 'unit', 'real_price': 69, 'direct_labor': 23, 'direct_wages': 54, 'indirect_wages': 12,
            'indirect_labor': 23
        })

    def test_create_product(self):
        #Post request to create new product
        res = self.client.post("/api/create/product", {
            'name': 'HotDog', 'measurement': 'unit', 'real_price': 69, 'direct_labor': 36, 'direct_wages': 50, 'indirect_wages': 12, 'indirect_labor': 23
        })
        self.assertEqual(res.status_code, 302)

    def test_get_product(self):
        #Get request for created product HotDog
        res = self.client.get('/product/borgor/')
        self.assertEqual(res.status_code, 200)

    def test_get_analytics(self):
        #get request for HotDog analytics
        res = self.client.get('/product/borgor/analytics/')
        self.assertEqual(res.status_code, 200)


from django.test import TestCase
from ..models import Product
from ..models import Dependency
from ..models import DependencyCycleError
from ..api import *


class CalcDirectTestCase(TestCase):
    def setUp(self):
        p1 = Product(name="Prod1", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p2 = Product(name="Prod2", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p1.save()
        p2.save()
        d = Dependency(dependent_id=p1.id, dependency_id=p2.id, quantity =10.0)
        d.save()

    def test_calculates_correct_values_without_indirection(self):
        update_product_indirect_values()
        self.assertEqual(Product.objects.get(id=1).indirect_labor, 100.0)
        self.assertEqual(Product.objects.get(id=2).indirect_labor, 0.0)

class CalcIndirTestCase(TestCase):
    def setUp(self):
        # TODO: Clean this stuff up
        p1 = Product(name="Prod1", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p2 = Product(name="Prod2", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p3 = Product(name="Prod3", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p4 = Product(name="Prod4", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p1.save()
        p2.save()
        p3.save()
        p4.save()
        d1 = Dependency(dependent_id=p2.id, dependency_id=p1.id, quantity =1.0)
        d2 = Dependency(dependent_id=p3.id, dependency_id=p1.id, quantity =1.0)
        d3 = Dependency(dependent_id=p3.id, dependency_id=p2.id, quantity =1.0)
        d4 = Dependency(dependent_id=p4.id, dependency_id=p3.id, quantity =1.0)
        d1.save()
        d2.save()
        d3.save()
        d4.save()

    def test_calculates_correct_values_with_indirection(self):
        update_product_indirect_values()
        self.assertEqual(Product.objects.get(id=1).indirect_labor, 0.0)
        self.assertEqual(Product.objects.get(id=2).indirect_labor, 10.0)
        self.assertEqual(Product.objects.get(id=3).indirect_labor, 30.0)
        self.assertEqual(Product.objects.get(id=4).indirect_labor, 40.0)

class CalcIndirCostTestCase(TestCase):
    def setUp(self):
        # TODO: Clean this stuff up
        p1 = Product(name="Prod1", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p2 = Product(name="Prod2", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p3 = Product(name="Prod3", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p4 = Product(name="Prod4", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p1.save()
        p2.save()
        p3.save()
        p4.save()
        d1 = Dependency(dependent_id=p2.id, dependency_id=p1.id, quantity =1.0)
        d2 = Dependency(dependent_id=p3.id, dependency_id=p1.id, quantity =1.0)
        d3 = Dependency(dependent_id=p3.id, dependency_id=p2.id, quantity =1.0)
        d4 = Dependency(dependent_id=p4.id, dependency_id=p3.id, quantity =1.0)
        d1.save()
        d2.save()
        d3.save()
        d4.save()

    def test_calculates_correct_values_with_indirection(self):
        update_product_indirect_values()
        self.assertEqual(Product.objects.get(id=1).indirect_wages, 0.0)
        self.assertEqual(Product.objects.get(id=2).indirect_wages, 10.0)
        self.assertEqual(Product.objects.get(id=3).indirect_wages, 30.0)
        self.assertEqual(Product.objects.get(id=4).indirect_wages, 40.0)

class DetectErrorCase(TestCase):
    def setUp(self):
        p1 = Product(name="Prod1", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p2 = Product(name="Prod2", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p3 = Product(name="Prod3", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p4 = Product(name="Prod4", measurement="unit", real_price=10.0, direct_labor=10.0, direct_wages=10.0)
        p1.save()
        p2.save()
        p3.save()
        p4.save()
        d1 = Dependency(dependent_id=p1.id, dependency_id=p1.id, quantity =1.0)
        d2 = Dependency(dependent_id=p2.id, dependency_id=p3.id, quantity =4.0)
        d3 = Dependency(dependent_id=p3.id, dependency_id=p2.id, quantity =0.5)
        d4 = Dependency(dependent_id=p4.id, dependency_id=p4.id, quantity =0.1)
        d1.save()
        d2.save()
        d3.save()
        d4.save()


    def test_detects_direct_error(self):
        update_product_indirect_values()
        self.assertIsNotNone(DependencyCycleError.objects.get(product_id=1))

    def test_detects_indirect_error(self):
        update_product_indirect_values()
        self.assertIsNotNone(DependencyCycleError.objects.get(product_id=2))

    def test_detects_no_error(self):
        update_product_indirect_values()
        try:
            DependencyCycleError.objects.get(product_id=4)
            self.fail("This object should not be possible to get.")
        except:
            pass



from django.shortcuts import render, redirect
from django.views.generic import TemplateView
from django.http import HttpResponseRedirect
from django.contrib.auth import logout as auth_logout
from .models import *
from .forms import *
from .decorators import *
import product_graph_bindings


def commit_history():
    """
    Commits the current graph state to history, for the purpose of analytics.
    """
    histpoint = HistoryPoint()
    histpoint.save()
    for p in Product.objects.all():
        p_history = ProductHistory(
            history_point=histpoint,
            product_id=p.id,
            name=p.name,
            measurement=p.measurement,
            real_price=p.real_price,
            direct_labor=p.direct_labor,
            direct_wages=p.direct_wages,
            indirect_wages=p.indirect_wages,
            indirect_labor=p.indirect_labor
        )
        p_history.save()
    for d in Dependency.objects.all():
        d_history = DependencyHistory(
            history_point=histpoint,
            dependent_id=d.dependent_id,
            dependency_id=d.dependency_id,
            quantity=d.quantity
        )
        d_history.save()


def commit_history_request(request, name=None):
    if request.method == "POST":
        commit_history()
    if name is None:
        return redirect("/products/")
    else:
        return redirect("/product/" + name)


def logout(request):
    """
    Logs a user out
    :param request: The request sent to server
    :type request: HttpRequest
    :return: a redirect to the home page on sucess, 404 if failure or request method is
        not POST
    :rtype: HttpResponseRedirect
    """
    if request.method == "POST":
        auth_logout(request)
        return redirect("/")

    return redirect("/fourohfour")


### CRUD FOR PRODUCT ###
@auth_required
def create_product(request):
    """
    Creates a product in the database. This function handles POST requests sent to the URL 'api/create/product'
    :param request: The request sent to server
    :type request: HttpRequest
    :return: a redirect to the homepage '/' on successful save to DB, to '/fourohfour' on failure or on request
        types that are not POST
    :rtype: HttpResponseRedirect
    """
    # handle the post to this url ONLY
    if request.method == 'POST':
        form = ProductForm(request.POST)
        if form.is_valid():
            product = Product(
                name=form.cleaned_data['name'],
                measurement=form.cleaned_data['measurement'],
                real_price=form.cleaned_data['real_price'],
                direct_labor=form.cleaned_data['direct_labor'],
                direct_wages=form.cleaned_data['direct_wages'],
            )

            product.save()
            # NOTE: Updating the indirect values here!
            update_product_indirect_values()
        else:
            print(form._errors)

        page = form.cleaned_data['page']

        if page is None or page == 1:
            return HttpResponseRedirect("/products")
        else:
            # do we want to route to the same page after creation?
            # should we route to the first / last page instead? If we
            # sort our list of products, this would be more difficult.
            # if the products are in order of creation, we could redirect to the
            # last page maybe?
            return HttpResponseRedirect("/products/?page={}".format(page))


    # redirect to 404 if method isn't post
    else:
        return HttpResponseRedirect("/fourohfour")


# TODO: add safety try/except blocks (see delete_product)
@auth_required
def edit_product(request, name):
    """
    Updates a product in the database. This function handles POST requests sent to the URL 'api/edit/product/:id'
    :param request: The request sent to server
    :type request: HttpRequest
    :param name: the name of the product selected to update
    :type name: str (This will change to int to reflect the future change in the Product model)
    :return: a redirect to the selected product info page using new id '/product/:id' on successful find and update to DB,
        to '/product/:id' using original id on failure and to '/fourohfour' on request types that are not POST
    :rtype: HttpResponseRedirect
    """
    # url should only accept post requests
    if request.method == 'POST':
        form = ProductForm(request.POST)
        if form.is_valid():
            Product.objects.filter(name=name).update(
                name=form.cleaned_data['name'],
                measurement=form.cleaned_data['measurement'],
                real_price=form.cleaned_data['real_price'],
                direct_labor=form.cleaned_data['direct_labor'],
                direct_wages=form.cleaned_data['direct_wages'],
            )
            # NOTE: Updating the indirect values here!
            update_product_indirect_values()
            # redirect using NEW name, since it may have been updated
            return HttpResponseRedirect("/product/{}".format(form.cleaned_data['name']))

        else:
            print(form._errors)

            # redirect to the product page using ORIGINAL name, since update did not work
            return HttpResponseRedirect("/product/{}".format(name))

    else:
        return HttpResponseRedirect("/fourohfour")


@auth_required
def delete_product(request, name):
    """
    Deletes a product in the database. This function handles POST requests sent to the URL 'api/delete/product/:id'
    :param request: The request sent to server
    :type request: HttpRequest
    :param name: the name of the product selected to delete
    :type name: str (This will change to int to reflect the future change in the Product model)
    :return: a redirect to the homepage '/' on successful find and delete to DB, to '/fourohfour' on failure or on request
        types that are not POST
    :rtype: HttpResponseRedirect
    """
    # TODO: change to DELETE request??
    if request.method == 'POST':
        try:
            # find by name (primary key)
            # if not found, goes to except block
            # delete on find
            Product.objects.get(name=name).delete()
            update_product_indirect_values()
        except:
            # TODO: is this the best action to take?
            # I think we should look into handling errors and notifying the user,
            # rather than routing to 404 anytime a DB operation fails
            return HttpResponseRedirect("/fourohfour")

        return HttpResponseRedirect("/products")
    else:
        return HttpResponseRedirect("/fourohfour")


### CRUD FOR DEPENDENCY ###
@auth_required
def create_dependency(request, prod_name):
    # handle the post to this url ONLY
    if request.method == 'POST':
        form = DependencyForm(request.POST)
        if form.is_valid():

            try:
                dependent = Product.objects.get(name=prod_name)
                dependency = Product.objects.get(name=form.cleaned_data['dependency'])

            except:
                return HttpResponseRedirect("/fourohfour")

            # Check if a dependency of this directionbetween these Products already exists
            if (Dependency.objects.filter(dependent=dependent).filter(dependency=dependency)):
                print("This dependency already exists!")
                return HttpResponseRedirect("/product/{}".format(prod_name))

            newDependency = Dependency(
                dependent=dependent,
                dependency=dependency,
                quantity=form.cleaned_data['quantity']

            )
            newDependency.save()
            update_product_indirect_values()
        else:
            print(form._errors)

        return HttpResponseRedirect("/product/{}".format(prod_name))

    # redirect to 404 if method isn't post
    else:
        return HttpResponseRedirect("/fourohfour")


@auth_required
def edit_dependency(request, prod_name):
    # url should only accept post requests
    if request.method == 'POST':
        form = EditDependencyForm(request.POST)
        if form.is_valid():
            try:
                dep = Dependency.objects.filter(id=form.cleaned_data['id'])
                dep.update(
                    dependent=Product.objects.get(name=prod_name),
                    # Assuming that the dependencies are dependent on the product
                    dependency=Product.objects.get(name=form.cleaned_data['dependency']),
                    quantity=form.cleaned_data['quantity']
                )
            except:
                print("Problem setting dep")
                return HttpResponseRedirect("/product/{}".format(prod_name))

            update_product_indirect_values()
            # redirect using NEW dependency, since it may have been updated
            return HttpResponseRedirect("/product/{}".format(prod_name))

        else:
            print("Error resulting from form", form._errors)

            # redirect to the dependency page using ORIGINAL name, since update did not work if here
            return HttpResponseRedirect("/product/{}".format(prod_name))

    else:
        return HttpResponseRedirect("/fourohfour")


@auth_required
def delete_dependency(request):
    # TODO: change to DELETE request??
    if request.method == 'POST':
        redirect_to = ""
        form = DeleteDependencyForm(request.POST)
        if form.is_valid():
            try:
                # find by id (primary key)
                # if not found, goes to except block
                # delete on find
                dep_id = form.cleaned_data['id']
                redirect_to = form.cleaned_data['redirect_to']
                Dependency.objects.get(id=dep_id).delete()
            except:
                # TODO: is this the best action to take?
                return HttpResponseRedirect("/fourohfour")

        else:
            print(form._errors)
            # TODO: maybe add routing to uh oh error page??

        update_product_indirect_values()
        return HttpResponseRedirect("/product/{}".format(redirect_to))
    else:
        return HttpResponseRedirect("/fourohfour")


### Calculating indirect costs ###

def update_product_indirect_values():
    """
    Performs the calculation of indirect values by calling on the Rust library.
    """
    # TODO: might want to make a separate function that handles a request, if we stop doing this
    # calculation automatically when data is added
    #if request.method != 'PUT':
    #    return HttpResponseRedirect("/fourohofur")

    # NOTE: this code has very bad perf and does many queries when maybe it can do group_by and stuff?
    # need to combine some loops at very least

    labor_graph = {}
    cost_graph = {}
    for p in Product.objects.all():
        deps = []
        for d in Dependency.objects.filter(dependent=p.id):
            deps.append((d.dependency_id, d.quantity))
        # Calculating labor time
        labor_graph[p.id] = product_graph_bindings.SimpleProduct(p.direct_labor, deps)

        # Calculating cost-price
        cost_graph[p.id] = product_graph_bindings.SimpleProduct(p.direct_wages, deps)

    (indirect_labor_values, errors) = product_graph_bindings.calc_indirect_vals_for_n_iterations(labor_graph, 25)

    # Check for errors in graph
    if len(errors) != 0:
        # clear old errors
        DependencyCycleError.objects.all().delete()
        # NOTE: batch inserts could be good here but they have some weird caveats
        for id in errors:
            newError = DependencyCycleError(product_id = id)
            newError.save()
        return
    else:
        # Clear any preexisting error data since the graph is now valid
        DependencyCycleError.objects.all().delete()

    (indirect_cost_values, errors) = product_graph_bindings.calc_indirect_vals_for_n_iterations(cost_graph, 25)

    # update products with calculated values
    for (id_val, indirect_labor_val) in indirect_labor_values:
        Product.objects.filter(id=id_val).update(indirect_labor=indirect_labor_val)



    # update products with calculated values
    for (id_val, indirect_cost_val) in indirect_cost_values:
        Product.objects.filter(id=id_val).update(indirect_wages=indirect_cost_val)



from django.db import models
from ..models import Product
from django.test import TestCase
class TestProduct(TestCase):
    def test_product_name(self):
        prod = Product(name="Hot Dog")
        self.assertEqual("Hot Dog", prod.name)
        prod = Product(name="This name is over 100 characters and should not work according to the model specs. We will obviously find out and see if that is the case.")
        self.assertEqual("This name is over 100 characters and should not work according to the model specs. We will obviously find out and see if that is the case.",prod.name)
    def test_product_measurement(self):
        prod = Product(measurement="Pound")
        self.assertEqual("Pound", prod.measurement)
        prod = Product(measurement="This measurement is over 40 characters and should not work.")



# Generated by Django 3.0.7 on 2020-06-30 03:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Economy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('real_average_profit_rate', models.FloatField()),
                ('estimated_average_profit_rate', models.FloatField()),
            ],
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
                ('real_price', models.FloatField()),
                ('direct_labor', models.FloatField()),
                ('direct_wages', models.FloatField()),
                ('indirect_wages', models.FloatField(default=0)),
                ('indirect_labor', models.FloatField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Dependency',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.FloatField()),
                ('dependency', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependencies', to='backend.Product')),
                ('dependent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependents', to='backend.Product')),
            ],
        ),
    ]



from django.test import TestCase
from .. import views
from ..models import Product
from django.test import Client

class TestViews(TestCase):
    def test_retrieve_product(self):
        prod = Product(name="Hot Dog")
        self.assertNotEqual(prod, views.retrieve_product("Bad Dog"))



#!/usr/bin/env python
import os
import sys
from django.core.management import call_command


def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'backend.settings')
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError("Couldn't import Django.") from exc
    
    # run server
    execute_from_command_line(["manage.py", "runserver"])
    

if __name__ == '__main__':
    main()



from django.shortcuts import render,redirect
from django.http import HttpResponse

def unauthed_route(view_fn):
    """
    Decorator to control routes only meant for unauthenticated users (e.g. registered users 
    don't need access to the login or register page)

    :param view_fn: The view function returned on sucess (unauthed user)
    :type request: Function

    :return: a redirect to the product list page '/products/' if user in request is authenticated, 
        a Function to render the requested template if user is unauthenticated
    :rtype: HttpResponseRedirect or Function
    """
    def wrapper_fn(request,*args, **kwargs):
        if request.user.is_authenticated:
            return redirect('/products/')
        else:
            return view_fn(request, *args, **kwargs)
    
    return wrapper_fn

def auth_required(api_fn):
    """
    Decorator to control routes only meant for authenticated users (e.g. creating products 
    / editing products requires a logged in user)

    :param view_fn: The API function returned on sucess (authed user)
    :type request: Function

    :return: a redirect to the product list page '/login/' if user in request is unauthenticated, 
        a Function to render the requested template if user is authenticated
    :rtype: HttpResponseRedirect or Function
    """
    def wrapper_fn(request, *args, **kwargs):
        if request.user.is_authenticated:
            return api_fn(request, *args, **kwargs)

        else:
            return redirect('/login/')
    
    return wrapper_fn


"""backend URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import path, include
from . import views
from . import api

urlpatterns = [
    path('admin/', admin.site.urls),
    path('fourohfour/', views.fourohfour),

    # user pages
    path('', views.home),
    path('login/', views.login, name="login"),
    path('register/', views.register),
    path('logout/', api.logout),

    # product pages
    path('products/', views.products_page),
    path('errors/', views.errors_page),
    path('products/?page=<int:num>', views.products_page),
    path('products/analytics', views.products_analytics),
    path('product/<str:name>/', views.product_view),
    path('product/<str:name>/analytics/', views.product_analytics),

    ### api ###
    path('api/create/product', api.create_product),
    path('api/edit/product/<str:name>', api.edit_product),
    path('api/delete/product/<str:name>', api.delete_product),
    path('api/create/dependency/<str:prod_name>', api.create_dependency),
    path('api/edit/dependency/<str:prod_name>', api.edit_dependency),
    path('api/delete/dependency/', api.delete_dependency),
    path('api/edit/commit-history/<str:name>', api.commit_history_request),
    path('api/edit/commit-history/', api.commit_history_request)
]



# Generated by Django 3.0.7 on 2020-07-31 20:02

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('backend', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoryPoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_time', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='product',
            name='measurement',
            field=models.CharField(default='Unit', max_length=40),
            preserve_default=False,
        ),
        migrations.CreateModel(
            name='ProductHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('product_id', models.IntegerField()),
                ('name', models.CharField(max_length=100)),
                ('measurement', models.CharField(max_length=40)),
                ('real_price', models.FloatField()),
                ('direct_labor', models.FloatField()),
                ('direct_wages', models.FloatField()),
                ('indirect_wages', models.FloatField(default=0)),
                ('indirect_labor', models.FloatField(default=0)),
                ('history_point', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='backend.HistoryPoint')),
            ],
        ),
        migrations.CreateModel(
            name='DependencyHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dependent_id', models.IntegerField()),
                ('dependency_id', models.IntegerField()),
                ('quantity', models.FloatField()),
                ('history_point', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='backend.HistoryPoint')),
            ],
        ),
        migrations.CreateModel(
            name='DependencyCycleError',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='backend.Product')),
            ],
        ),
    ]



import scripts
if __name__ == "__main__":
    import setuptools
    setuptools.setup(
        name='productiongraph',
        version='0.2.0',
        author='Cameron, Aaron, Osiris, Benjamin, Matt',
        author_email='cameron.g.rodgers@gmail.com',
        packages=setuptools.find_packages(),
        #include_package_data=True,
        package_data={
            "": ["*.html", "*.css", "*.jpg", "*.png"],
            "productiongraph": ["*.html", "*.css", "*.jpg", "*.png"],
            "backend": ["*.html", "*.css", "*.jpg", "*.png"],
        },
        description='Web interface for calculating prices',
        install_requires=[ 'Django', 'product_graph_bindings', 'numpy', 'django_filter' ],
        entry_points =
        { 'console_scripts':
            [
                'prodgraph_run = scripts.run:main',
                'prodgraph_migrate_and_run = scripts.migrate_and_run:main',
                'prodgraph_init = scripts.init:main',
                'prodgraph_test = scripts.test:main',
            ]
        },
    )



from django import forms
from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth.models import User

class CreateUserForm(UserCreationForm):
    class Meta:
        model = User
        fields = [
            'username',
            'password1',
            'password2',
        ]

class ProductForm(forms.Form):
    name = forms.CharField(max_length = 100) # NOTE: this max length is arbitrary placeholder
    measurement = forms.CharField(max_length = 40)
    real_price = forms.FloatField()
    direct_labor = forms.FloatField() # Consider placing a lower bound of 0.0?
    direct_wages = forms.FloatField()
    page = forms.IntegerField(required = False)

class DependencyForm(forms.Form):
    # The Dependent is not needed in the form, since it can be gotten from product.name in the html file
    dependency = forms.CharField()
    quantity = forms.FloatField()

class EditDependencyForm(forms.Form):
    # The Dependent is not needed in the form, since it can be gotten from product.name in the html file
    id = forms.FloatField()
    dependency = forms.CharField()
    quantity = forms.FloatField()

class DeleteDependencyForm(forms.Form):
    id = forms.FloatField()
    redirect_to = forms.CharField(max_length = 100)